#!/usr/bin/python2
#  $Id: mirror-dist.py,v 1.15 2005/04/20 07:09:43 bharring Exp $
#
#  Copyright (C) 2005
#
#  Brian Harring
#  ferringb@gentoo.org
#  see Copying in mirror-dist tarball for GPL-2 rules.

import atexit
import optparse
import os
import Queue
import re
import shelve
import shutil
import signal
import stat
import sys
import threading
import time
import traceback
import warnings

import fetching_lib

import portage
import portage.data as portage_data
import portage.exception as portage_exception
import portage.util as portage_util
import portage.versions as portage_versions

if sys.hexversion >= 0x3000000:
	long = int

hash_name = portage.const.MANIFEST2_REQUIRED_HASH
hash_func = portage.checksum.hashfunc_map[hash_name]
conditional_re = re.compile(r'\S+\?\s+\(.*\)')
default_fetch_string = "wget -c -v -t 1 --passive-ftp --timeout=60 -O %s %s"

available_options = (
	{
		"longopt"  : "--failure-log",
		"shortopt" : "-f",
		"help"     : "specific file to log failures in",
	},
	{
		"longopt"  : "--success-log",
		"shortopt" : "-s",
		"help"     : "specific file to log success in",
	},
	{
		"longopt"  : "--general-log",
		"shortopt" : "-g",
		"help"     : "specific file to log general/remaining output in",
	},
	{
		"longopt"  : "--append",
		"shortopt" : "-a",
		"help"     : "append to logs, rather than wipe",
		"action"   : "store_true"
	},
	{
		"longopt"  : "--cleanse-distfiles",
		"shortopt" : "-c",
		"help"     : "enable distfiles cleansing",
		"action"   : "store_true"
	},
	{
		"longopt"  : "--death-list",
		"help"     : "file to store a list of files slated for death",
	},
	{
		"longopt"  : "--death-watch-seconds",
		"help"     : "period of time on death list, prior to removal",
	},
	{
		"longopt"  : "--porttree",
		"shortopt" : "-p",
		"help"     : "specific porttree to use",
	},
	{
		"longopt"  : "--temp-location",
		"shortopt" : "-t",
		"help"     : "temporary location to download to",
	},
	{
		"longopt"  : "--distfiles",
		"shortopt" : "-d",
		"help"     : "distfiles directory to use",
	},
	{
		"longopt"  : "--mirror-overrides",
		"shortopt" : "-m",
		"help"     : "file holding a list of mirror overrides",
	},
	{
		"longopt"  : "--mirror-skip",
		"help"     : "comma delimited list of mirror targets to skip "
			"when fetching",
	},
	{
		"longopt"  : "--nomirror-exemptions",
		"help"     : "comma delimited list of mirror targets to ignore "
			"RESTRICT=\"mirror\" for",
	},
	{
		"longopt"  : "--threads",
		"help"     : "# of fetcher threads to use",
	},
	{
		"longopt"  : "--verify-existing-digest",
		"help"     : "use digest as a verification of whether existing "
			"distfiles are valid",
		"action"   : "store_true"
	},
	{
		"longopt"  : "--verify-existing-md5",
		"help"     : "deprecated alias for --verify-existing-digest",
		"action"   : "store_true"
	},
	{
		"longopt"  : "--verify-existing-size",
		"help"     : "use size as a verification of whether existing "
			"distfiles are valid",
		"action"   : "store_true"
	},
	{
		"longopt"  : "--debug",
		"help"     : "debug mode.  don't actually fetch, nor delete",
		"action"   : "store_true"
	},
	{
		"longopt"  : "--distfiles-local",
		"help"     : "distfiles-local directory to use",
	},
	{
		"longopt"  : "--distfiles-tracking",
		"help"     : "track which ebuilds a distfile belongs to",
	},
	{
		"longopt"  : "--max-attempts",
		"help"     : "maximum number of tries per file",
	},
	{
		"longopt"  : "--disable-central-pid-manager",
		"help"     : "If wget is going zombie, this should make it work",
		"action"   : "store_true"
	},
	{
		"longopt"  : "--purgatory-dir",
		"help"     : "instead of removing stale distfiles, transfer them "
			"to a directory. This directory is checked when "
			"trying to pull files",
	},
	{
		"longopt"  : "--purgatory-db",
		"help"     : "specify location of a purgatory db for tracking files",
	},
	{
		"longopt"  : "--purgatory-death-watch-seconds",
		"help"     : "how long a file can live in purgatory.  If unset and "
			"purgatory-db is on, defaults to 180 days",
	},
	{
		"longopt"  : "--individual-fetch-logsdir",
		"help"     : "directory to dump fetch logs from each attempted fetch",
	},
	{
		"longopt"  : "--whitelist",
		"help"     : "list of files to whitelist, one per line, # prefixed "
			"lines ignored",
	},
)

fail_log = None
suc_log = None
gen_log = None
fail_lock = threading.Lock()
suc_lock  = threading.Lock()
gen_lock  = threading.Lock()
failed_thread_kill_switch = threading.Event()
worker_kill_switch = threading.Event()

debug_p=0
def dprint(s):
	global debug_p
	if debug_p > 1:
		print s

def fail_print(msg):
	fail_lock.acquire()
	try:	fail_log.write(msg)
	except (OSError, IOError):	print "failure writing to failure log"
	fail_lock.release()

def suc_print(msg):
	suc_lock.acquire()
	try:	suc_log.write(msg)
	except (OSError, IOError):	print "failure writing to success log"
	suc_lock.release()

def gen_print(msg):
	gen_lock.acquire()
	try:	gen_log.write(msg)
	except (OSError, IOError):
		try:	print "failure writing to general log"
		except (OSError, IOError):
			# time to kill this motha.
			raise RuntimeError
	gen_lock.release()

InvalidDependString = portage_exception.InvalidDependString
pkgsplit = portage_versions.pkgsplit
catpkgsplit = portage_versions.catpkgsplit
grabdict = portage_util.grabdict
movefile = portage.movefile

if os.getuid() == 0:
	uid = portage_data.portage_uid
	gid = portage_data.portage_gid
else:
	uid = os.getuid()
	gid = os.getgid()

def file_generator(dbapi, mirror_callable, kill_switch, file_list={}, nomirror_exemptions=[]):
	"""dbapi must be a cp_list/cp_all/aux_get compatible, mirror_callable being a mangler of uris for 
	mirror:// uri, file_list is if the callers desires a dict of all files"""

	for cp in dbapi.cp_all():

		if kill_switch.isSet():		return

		# Reset state so the Manifest is pulled once for this cp.
		digests = None

		for cpv in dbapi.cp_list(cp):
			# why do we get the uri ourselves?  because fetchlist has a sys.exit in it.
			#
			# that is a very _stupid_ place for a sys.exit.
			#
			# hence, we reinvent the wheel.
			#
			try:
				restrict, = dbapi.aux_get(cpv, ("RESTRICT",))
			except SystemExit:
				raise
			except Exception, e:
				fail_print("%s\t\taux_get exception %s" % (cpv, e))
				del e
				continue

			# Here we ignore conditional parts of RESTRICT since they don't
			# apply unconditionally. Assume such conditionals only apply on
			# the client side.
			restrict = conditional_re.sub('', ' '.join(restrict.split())).split()

			if filter(lambda x: "fetch" in x, restrict):
				gen_print("%s: skipping, fetch restriction\n" % cpv)
				continue

			try:
				uri_map = dbapi.getFetchMap(cpv)
			except SystemExit:	raise
			except Exception, e:
				fail_print("%s\t\tgetFetchMap exception %s\n" % (cpv, e))
				del e
				continue

			if not uri_map:
				#gen_print("%s: skipping, no uris\n" % cpv)
				continue

			if filter(lambda x: "mirror" in x, restrict):
				skip = False
				if nomirror_exemptions:
					new_uri_map = {}
					for filename, uri_set in uri_map.iteritems():
						for uri in uri_set:
							if uri[:9] == "mirror://":
								i = uri.find("/", 9)
								if i != -1 and uri[9:i] in nomirror_exemptions:
									new_uri_map[filename] = uri_set
									break
					if new_uri_map:
						uri_map = new_uri_map
					else:
						skip = True
				else:
					skip = True

				if skip:
					gen_print("%s: skipping, mirror restriction and no exempted uris\n" % cpv)
					continue

			# There's only one Manifest per cp, and digestParseFile() returns
			# all digests in the Manifest (since portage-2.1).
			if digests is None:
				try:
					digests = portage.Manifest(
						os.path.join(dbapi.porttree_root, cp),
						None).getTypeDigests("DIST")
				except SystemExit:
					raise
				except Exception, e:
					for filename in uri_map:
						fail_print("%s\t%s\tManifest exception %s\n" % \
							(cpv, filename, e))
					del e
					continue

			if not digests:
				for filename in uri_map:
					fail_print("%s\t%s\tdigest entry missing\n" % (cpv, filename))
				continue

			for filename, uri_set in uri_map.iteritems():
				if filename in file_list:
					#onward, wayward son.
					continue
				if filename not in digests:
					fail_print("%s\t%s\tdigest entry missing\n" % (cpv, filename))
					continue
				file_list[filename] = cpv
				d = {"cpv":cpv}
				d.update(digests[filename])

				try:
					yield fetching_lib.file_to_fetch(
						filename, mirror_callable(filename, uri_set), extra=d)
				except SystemExit: raise
				except Exception, e:
					fail_print("%s\t%s\texception: %s\n" % \
						(cpv, filename, e))
					del e
					continue

class digest_verifier(object):
	def __init__(self, location, func):	
		self.func = func
		self.location = location
	def __call__(self, file):
		return self.func(os.path.join(self.location,file.filename), 
			file.extra_data.get("size",0L), 
			file.extra_data.get(hash_name, ''))

class size_verifier(object):
	def __init__(self, location):	self.location = location
	def __call__(self, file):
		try:	st=os.stat(os.path.join(self.location,file.filename))
		except (OSError, IOError):
			return False
		return st.st_size == file.extra_data.get("size",0L)


class failed_processor(threading.Thread):
	def __init__(self, queue, kill_switch, logger, timeout=10):
		self.queue = queue
		self.kill_switch = kill_switch
		self.logger = logger
		self.timeout = timeout
		self.failures = 0
		threading.Thread.__init__(self)

	def run(self):
		while not self.kill_switch.isSet():
			try:			entry = self.queue.get(1, self.timeout)
			except Queue.Empty:	continue
			self.failures += 1
			file_inst, err = entry
			cpv = file_inst.extra_data.get("cpv","not specified")
			if isinstance(err, Exception):
				self.logger("%s\t%s\texception %s\n" % \
					(cpv, file_inst.filename, str(err)))
			else:
				self.logger("%s\t%s\t%s\n" % \
					(cpv, file_inst.filename, str(err)))

class ShelveUnicodeWrapper(object):
	"""
	Convert unicode to str and back again, since python-2.x shelve
	module doesn't support unicode.
	"""
	def __init__(self, shelve_instance):
		self.__data = shelve_instance

	def _encode(self, s):
		if isinstance(s, unicode):
			s = s.encode('utf_8')
		return s

	def __contains__(self, k):
		return self._encode(k) in self.__data

	def __setitem__(self, k, v):
		self.__data[self._encode(k)] = self._encode(v)

	def __getitem__(self, k):
		return self.__data[self._encode(k)]

	def __delitem__(self, k):
		del self.__data[self._encode(k)]

	def get(self, k, *args):
		return self.__data.get(self._encode(k), *args)

	def keys(self):
		return self.__data.keys()

	def iterkeys(self):
		return self.__data.iterkeys()

	def close(self):
		self.__data.close()

	def clear(self):
		self.__data.clear()

if __name__=="__main__":

	death_int = None
	death_list = None
	tmp_location = None
	cleaning = False
	mirror_overrides_file = None;	mirror_override_dict = {}
	open_mode = "w"
	distdir = None
	distfiles_local = None
	thread_count = 5
	cleansing = False
	verify="size"
	debug = False
	url_db = None
	max_tries = 10
	run_cpidw = True
	purgatory_dir = None
	purgatory_db = None
	purgatory_int = 180*24*3600
	indie_fetch_dir = None
	whitelist = {}
	nomirror_exemptions = []
	mirror_skip = []

	parser = optparse.OptionParser()
	for opt_info in available_options:
		opt_pargs = [opt_info["longopt"]]
		if opt_info.get("shortopt"):
			opt_pargs.append(opt_info["shortopt"])
		opt_kwargs = {"help" : opt_info["help"]}
		if opt_info.get("action"):
			opt_kwargs["action"] = opt_info["action"]
		parser.add_option(*opt_pargs, **opt_kwargs)

	options, args = parser.parse_args(args=sys.argv[1:])

	if args:
		parser.error("remaining args: %s" % " ".join(args))

	if options.failure_log is not None:
		fail_log = os.path.abspath(options.failure_log)

	if options.success_log is not None:
		suc_log = os.path.abspath(options.success_log)

	if options.general_log is not None:
		gen_log = os.path.abspath(options.general_log)

	if options.cleanse_distfiles:
		cleansing = True

	if options.death_list is not None:
		death_list = os.path.abspath(options.death_list)

	if options.append:
		open_mode = "a"

	if options.death_watch_seconds is not None:
		death_int = long(options.death_watch_seconds)

	if options.temp_location is not None:
		tmp_location = os.path.abspath(options.temp_location)

	if options.distfiles is not None:
		distdir = os.path.abspath(options.distfiles)

	if options.mirror_overrides is not None:
		mirror_overrides_file = os.path.abspath(options.mirror_overrides)

	if options.mirror_skip is not None:
		mirror_skip = options.mirror_skip.split(",")

	if options.threads is not None:
		thread_count = int(options.threads)

	if options.verify_existing_digest:
		verify = "digest"

	if options.verify_existing_md5:
		verify = "digest"
		warnings.warn("--verify-existing-md5 is deprecated, please "
			"use --verify-existing-digest instead",
			UserWarning, stacklevel=0)

	if options.verify_existing_size:
		verify = "size"

	if options.debug:
		debug = True

	if options.distfiles_local is not None:
		distfiles_local = os.path.abspath(options.distfiles_local)

	if options.distfiles_tracking is not None:
		url_db = os.path.abspath(options.distfiles_tracking)

	if options.max_attempts is not None:
		max_tries = int(options.max_attempts)

	if options.disable_central_pid_manager:
		run_cpidw = False

	if options.purgatory_dir is not None:
		purgatory_dir = os.path.abspath(options.purgatory_dir)

	if options.purgatory_db is not None:
		purgatory_db = os.path.abspath(options.purgatory_db)

	if options.purgatory_death_watch_seconds is not None:
		purgatory_int = long(options.purgatory_death_watch_seconds)

	if options.individual_fetch_logsdir is not None:
		indie_fetch_dir = os.path.abspath(options.individual_fetch_logsdir)

	if options.whitelist is not None:
		whitelist = os.path.abspath(options.whitelist)

	if options.nomirror_exemptions is not None:
		nomirror_exemptions = options.nomirror_exemptions.split(",")

	os.environ["PORTDIR_OVERLAY"] = ""
	if options.porttree is not None:
		if not options.porttree.startswith(os.path.sep):
			options.porttree = os.path.join(os.getcwd(), options.porttree)
		os.environ["PORTDIR"] = options.porttree

	porttree = portage.settings["PORTDIR"]
	if not os.path.isdir(porttree) or not os.access(porttree, os.R_OK):
		print "porttree '%s' is not a dir, or not read accessible!" % porttree
		sys.exit(1)

	dbapi = portage.portdb

	if distdir == None:	distdir = portage.settings["DISTDIR"]
	else:
		if not distdir.startswith(os.path.sep):
			distdir = os.path.join(os.getcwd(), distdir)


	if debug:
		#debugging.
		print "debugging mode active.  won't fetch anything, nor delete."
		print "will have the same output however."
		print "any death-list passed in, will be updated."

	if not os.path.isdir(distdir) or not os.access(distdir, os.W_OK):
		print "distdir '%s' is not a dir, or not write accessible!" % distdir
		sys.exit(1)
	distdir = os.path.normpath("///"+distdir.rstrip(os.path.sep))

	if tmp_location != None:
		if not tmp_location.startswith(os.path.sep):
			tmp_location = os.path.join(os.getcwd(), tmp_location)

		if not os.path.isdir(tmp_location) or not os.access(tmp_location, os.W_OK):
			print "temp-location '%s' is not a dir, or not write accessible!" % tmp_location
			sys.exit(1)

		tmp_location = os.path.normpath("///"+tmp_location.rstrip(os.path.sep))
	else:
		tmp_location = distdir

	if purgatory_dir:
		if not purgatory_dir.startswith(os.path.sep):
			purgatory_dir = os.path.join(os.getcwd(), purgatory_dir)

		if  not (os.path.isdir(purgatory_dir) and os.access(purgatory_dir, os.W_OK)):
			print "remove-to-dir '%s' is either not a dir, or not write accessible!" % purgatory_dir
			sys.exit(1)
		purgatory_dir = os.path.normpath("///"+purgatory_dir.rstrip(os.path.sep))

	if purgatory_db:
		if not purgatory_dir:
			print "purgatory-db requires purgatory dir to be specified!"
			sys.exit(1)
		try:	purgatory_db = shelve.open(purgatory_db,"c")
		except (OSError, IOError), e:
			print "failed opening/creating purgatory db %s with exception %s" % (purgatory_db, e)
		else:
			if sys.hexversion < 0x3000000:
				purgatory_db = ShelveUnicodeWrapper(purgatory_db)

	if indie_fetch_dir:
		if not indie_fetch_dir.startswith(os.path.sep):
			indie_fetch_dir = os.path.join(os.getcwd(), indie_fetch_dir)

		if  not (os.path.isdir(indie_fetch_dir) and os.access(indie_fetch_dir, os.W_OK)):
			print "remove-to-dir '%s' is either not a dir, or not write accessible!" % indie_fetch_dir
			sys.exit(1)
		indie_fetch_dir = os.path.normpath("///"+indie_fetch_dir.rstrip(os.path.sep))

	if isinstance(whitelist, str):
		if not os.access(whitelist, os.R_OK):
			print "whitelist %s wasn't read accessible" % whitelist
			sys.exit(1)
		f=file(whitelist,"r")
		whitelist={}
		for x in f:
			x=x.lstrip().rstrip()
			if len(x) == 0 or x.startswith("#"):	continue
			whitelist[x] = True
		f.close()

	if mirror_overrides_file != None:
		if not os.path.isfile(mirror_overrides_file) or not os.access(mirror_overrides_file,os.R_OK):
			print "mirror-overrides-file '%s' is daft, either not a file, or not readable" % \
				mirror_overrides_file
			sys.exit(1)
		mirror_override_dict = grabdict(mirror_overrides_file)

	if thread_count == 0:
		print "thread count can't be set to 0.  resetting it to the 1"
		thread_count = 1

	if url_db != None:
		try:	url_db = shelve.open(url_db,"c")
		except (OSError, IOError), e:
			print "failed opening/creating url db %s with exception %s" % (url_db, e)
		else:
			if sys.hexversion < 0x3000000:
				url_db = ShelveUnicodeWrapper(url_db)

	death_db = {}
	if cleansing:
		if (death_int != None and death_list == None) or (death_int == None and death_list != None):
			parser.error("sorry, --death-list and --death-watch-seconds require each other.")

		if death_list:
			try:	death_db = shelve.open(death_list,"c")
			except (OSError, IOError), e:
				print "failed creating death-list %s- %s" % (death_list, str(e))
				sys.exit(2)
			else:
				if sys.hexversion < 0x3000000:
					death_db = ShelveUnicodeWrapper(death_db)

	if gen_log:
		f = file(gen_log, open_mode, 0)
		sys.stdout.close()
		sys.stdout = f
		sys.stderr.close()
		sys.stderr = f
		gen_log = f
	else:
		gen_log = sys.stdout
		sys.stderr.close()
		sys.stderr = sys.stdout
	if fail_log:		fail_log = file(fail_log, open_mode, 0)
	else:			fail_log = sys.stdout
	if suc_log:		suc_log  = file(suc_log, open_mode, 0)
	else:			suc_log = sys.stdout

	# ready to rock.
	# set a default traceback handler.
	def exception_handler(type, value, tb):
		traceback.print_stack()
		exit_handler()
		gen_print("exception caught.  signaled threads to die, although it may take a while\n")

	sigged = 0
	exit_code = 0
	def exit_handler(*args):
		global failed_thread_kill_switch, worker_kill_switch, sigged, exit_code
		exit_code = 1
		if len(args):
			print "caught sig",args[0]
		failed_thread_kill_switch.set()
		worker_kill_switch.set()
		if sigged > 1:
			# print "second sig caught.  forcing death"
			try:
				os.kill(-os.getpid(), signal.SIGKILL)
			except OSError: pass
			try:	os.wait()
			except OSError: pass
		gen_print("shutting down\n")
		sigged += 1

	atexit.register(exit_handler)

	signal.signal(signal.SIGINT, exit_handler)

	import copy
	mirrors = copy.copy(portage.thirdpartymirrors)
	mirrors.update(mirror_override_dict)

	for x in mirror_skip:
		mirrors[x] = []

	mirror_translator = fetching_lib.mirror_uri_processor(mirrors, max_tries)
	chf_obj = fetching_lib.checksum_verifier(hash_func)

	def noman(*args,**kwargs):	return False
	if debug:
		def f(*args):	return True
		chf_verifier = f
		purgatory_verifier = noman
	else:
		chf_verifier = chf_obj.verify
		purgatory_verifier = noman

	# got our existing distfiles verifier.
	if verify == "digest":
		existing_verifier = digest_verifier(distdir, chf_obj.verify)
	else:
		existing_verifier = size_verifier(distdir)

	def hardlink_or_copy(src, dest, info):
		# return True if file is successfully hardlinked or copied
		try:
			os.link(src, dest)
		except OSError, e:
			if os.path.exists(dest):
				# file created by a concurrent process
				return False
			else:
				fail_print("%s hardlink failed: %s\n" % (info, e))
				try:
					shutil.copy2(src, dest)
				except EnvironmentError, e:
					fail_print("%s copy2 failed: %s\n" % (info, e))
					return False
		return True

	def do_distiles_local(file_info):
		# return True if a new hardlink or copy is successfully created

		dist_local_file = os.path.join(distfiles_local, file_info.filename)
		try:
			dist_local_st = os.stat(dist_local_file)
		except OSError:
			return False

		distdir_file = os.path.join(distdir, file_info.filename)
		try:
			distdir_st = os.stat(distdir_file)
		except OSError:
			distdir_st = None

		if distdir_st is not None and \
			distdir_st.st_ino == dist_local_st.st_ino and \
			distdir_st.st_dev == dist_local_st.st_dev:
			# already hardlinked
			return False

		if distdir_st is None:
			return hardlink_or_copy(dist_local_file, distdir_file,
				"%s in distfiles-local" % file_info.filename)

		if distdir_st.st_dev != dist_local_st.st_dev:
			# Can't do cross-device hardlink.
			return False

		if distdir_st.st_size != dist_local_st.st_size:
			# files are different
			return False

		try:
			dist_local_digest = hash_func(dist_local_file)
			distdir_digest = hash_func(distdir_file)
		except EnvironmentError:
			return False

		if dist_local_digest != distdir_digest:
			# files are different
			return False

		head, tail = os.path.split(distdir_file)
		hardlink_tmp = os.path.join(head, ".%s._mirrordist_hardlink_.%s" % \
			(tail, os.getpid()))

		try:
			try:
				os.link(dist_local_file, hardlink_tmp)
			except OSError, e:
				fail_print("%s in distfiles-local hardlink failed: %s\n" %
					(file_info.filename, e))
				return False

			try:
				os.rename(hardlink_tmp, distdir_file)
			except OSError, e:
				fail_print("%s in distfiles-local hardlink rename failed: %s\n" %
					(file_info.filename, e))
				return False
		finally:
			try:
				os.unlink(hardlink_tmp)
			except OSError:
				pass

		return True

	valid_files_dict={}
	# do this now.  why?  cause it'll implode prior to starting up the threads if the 
	# tree is daft/non-existant

	threads = []

	completed_queue = Queue.Queue()
	failed_queue = Queue.Queue()
	failed_thread_kill_switch.clear()
	worker_kill_switch.clear()

	uri_source = file_generator(dbapi, mirror_translator, worker_kill_switch, file_list=valid_files_dict, 
		nomirror_exemptions=nomirror_exemptions)
	cpidw = fetching_lib.CentralPidWatcher(worker_kill_switch)

	if debug:
		def yes_man(*args, **kwargs):	return True
		fetcher = yes_man
	else:
		fetcher = fetching_lib.fetcher(default_fetch_string, tmp_location, uid=uid, gid=gid, pid_manager=cpidw, \
			logging_dir=indie_fetch_dir)

	worker_process_switch = threading.Event()
	worker_process_switch.set()
	job_queue = fetching_lib.job_tracker(0.2, worker_process_switch, worker_kill_switch)

	# start up cpidw.
	if run_cpidw:
		cpidw.start()

	# start up the failed queue logger..
	failed_thread = failed_processor(failed_queue, failed_thread_kill_switch, fail_print, timeout=0.5)
	failed_thread.start()
	
	while len(threads) < thread_count:
		threads.append(fetching_lib.fetcher_thread(job_queue, completed_queue, failed_queue, fetcher))
		threads[-1].start()


	#note the startup time.
	start_time = time.time()
	new_byte_count = 0

	# threads are all up and running, waiting for input.
	for file in uri_source:
		dprint("in uri_source still")
		new_file = False
		if url_db != None:
			if file.filename not in url_db:
				if not isinstance(url_db, dict):
					new_file = True
				url_db[file.filename] = file.extra_data["cpv"]

		if existing_verifier(file):
			if new_file:
				size = file.extra_data.get("size")
				if size is not None:
					new_byte_count += size
					suc_print("%s\t%s\tadded %i bytes\n" %
						(file.extra_data["cpv"], file.filename, size))
			if distfiles_local:
				do_distiles_local(file)
		else:

			# If the file exists then it's invalid, so remove it.
			try:
				os.unlink(os.path.join(distdir, file.filename))
			except OSError:
				pass

			if not debug and purgatory_dir:
				ret = False
				pfile = os.path.join(purgatory_dir, file.filename)
				cleanse_purgatory = False
				if os.path.exists(pfile):
					# note if it exists in purgatory, it gets removed.
					# ret controls whether or not it's fetched.
					cleanse_purgatory = True
					if "size" not in file.extra_data or hash_name not in file.extra_data:
						fail_print("%s in purgatory has missing chksum, removing it\n" % file.filename)
					elif chf_verifier(pfile, file.extra_data["size"], file.extra_data[hash_name]):
						try:	
							ret=movefile(pfile, os.path.join(distdir, file.filename))
							gen_print("reused %s from purgatory, removing it from purgatory\n" \
								% file.filename)
						except SystemExit: raise
						except Exception: pass
					else:
						fail_print("%s in purgatory has wrong chksum, removing it\n" % file.filename)

				if cleanse_purgatory:
					# purgatory file sucks.  so we kill it.
					try:	os.unlink(pfile)
					except OSError, oe:
						fail_print("%s failed in removal, caught exception %s\n" \
							% (file.filename, str(oe)))
						del oe

				if not ret and distfiles_local and \
					do_distiles_local(file) and \
					"size" in file.extra_data and \
					hash_name in file.extra_data and \
					chf_verifier(os.path.join(distdir, file.filename),
					file.extra_data["size"], file.extra_data[hash_name]):
					ret = True
					if new_file:
						size = file.extra_data.get("size")
						if size is not None:
							new_byte_count += size
							suc_print("%s\t%s\tadded %i bytes\n" %
								(file.extra_data["cpv"], file.filename, size))

				if not ret:
					job_queue.publish(file)
			else:
				job_queue.publish(file)
		if worker_kill_switch.isSet():
			break

	dprint("done with uri_source")

	# by the time this is reached, all uri of the tree are ready to go.
	# so... what is the main thread to do, aside from screw off?
	# well.  if cleanse is on... >:)
	cleansed_files = 0
	death_watch_count = 0

	if url_db == None:
		url_db = {}

	if cleansing and not worker_kill_switch.isSet():
		dprint("startingup cleansing")
		death_db_updates = []
		fatality_list = []
		dprint("doing distdir listdir")
		for file in os.listdir(distdir):
			try:	st=os.stat(os.path.join(distdir,file))
			except (OSError, IOError), e:
				gen_print("failed stat'ing '%s'\n" % file)
				continue
			if not stat.S_ISREG(st.st_mode):
				pass
			elif file in valid_files_dict:
				if file in death_db:
					gen_print("file %s was moved off the death watch: claimed by '%s'\n" % (file, valid_files_dict[file]))
					del death_db[file]
			elif file in whitelist:
				if file in death_db:
					gen_print("file %s was moved off the death watch: whitelisted\n" % file)
					del death_db[file]
			elif distfiles_local and os.path.exists(
				os.path.join(distfiles_local, file)):
				if file in death_db:
					gen_print("file %s was moved off the death watch: distfiles-local\n" % file)
					del death_db[file]
			elif file in death_db:
				pass
			elif death_int != None:
				# we have a waiting period for fatality.
				gen_print("file %s was moved onto the death watch- last owner- '%s'\n" % (file, url_db.get(file,"unknown")))
				death_db_updates.append(file)
			else:
				fatality_list.append(file)

		dprint("finished distdir listdir")

		for key in list(death_db.keys()):
			if death_db[key] + death_int <= start_time:
				fatality_list.append(key)
				del death_db[key]
		
		for file in fatality_list:
			dprint("doing fatality_list")
			if not debug:
				#finish him!
				if purgatory_dir:
					ret=False
					try:	ret=movefile(os.path.join(distdir,file), os.path.join(purgatory_dir,file))
					except SystemExit:	raise
					except Exception, e:	ret=e
					if not ret:
						fail_print("%s\tdeletion to %s failed for %s: %s\n" %  \
							(url_db.get(file,"unknown"), purgatory_dir, file, str(ret)))
						del ret
						continue
				else:
					try:	os.unlink(os.path.join(distdir, file))
					except (OSError, IOError), e:
						fail_print("unknown cpv\tfailed to remove unknown distfile %s due to exception '%s'\n" \
							 % (file, str(e)))
						del e
						continue
			suc_print("%s\t%s\tremoved\n" % (url_db.get(file,"unknown") ,file))

		if not isinstance(url_db, dict):
			dprint("cleansing url_db")
			for file in fatality_list:
				try:	del url_db[file]
				except KeyError: pass
			
		cleansed_files = len(fatality_list)
		del fatality_list
		death_watch_count = len(death_db.keys()) + len(death_db_updates)
		if not isinstance(death_db, dict) and not debug:
			for x in death_db_updates:
				death_db[x] = start_time
			del death_db_updates
		else:
			death_db.clear()

	if not isinstance(death_db, dict):
		death_db.close()

	valid_files_dict.clear()
	if isinstance(url_db, dict):	url_db.clear()
	else:				url_db.close()
	
	if purgatory_db is not None:
		unseen = dict([(x, None) for x in purgatory_db.iterkeys()])
		for f in os.listdir(purgatory_dir):
			fp = os.path.join(purgatory_dir, f)
			try:
				st = os.lstat(fp)
			except OSError:
				continue
			if not stat.S_ISREG(st.st_mode):
				continue
			elif f not in unseen:
				# we use unseen since dict is faster then shelve lookup.
				purgatory_db[f] = (st.st_size, start_time)
			else:
				del unseen[f]
				if long(purgatory_db[f][0]) != st.st_size:
					purgatory_db[f] = (st.st_size, start_time)
				elif purgatory_db[f][1] + purgatory_int < start_time:
					try:
						os.unlink(fp)
					except (OSError, IOError):
						gen_print("WARNING\tfailed to unlink purgatory file %s!" % fp)
					else:
						del purgatory_db[f]

		# wiped files, now remove dead entries.
		for x in unseen:
			del purgatory_db[x]
		purgatory_db.close()

		del unseen

	dprint("starting completed processing")

	# at this point, we've done our death dealing.  now we do verification of new files, requeuing as needed.
	move_required = tmp_location != distdir

	while not (worker_kill_switch.isSet() or (job_queue.no_work_left(thread_count) and completed_queue.empty())):
		dprint("in job queue looping")
		try:			entry = completed_queue.get(1, 0.25)
		except Queue.Empty:	continue
		# if no jobs, then go into a polling mode via while loop and delay above.
		
		file = entry.filename
		size = entry.extra_data.get("size","not specified")
		digest = entry.extra_data.get(hash_name, "")
		cpv = entry.extra_data.get("cpv","not specified")

		if not chf_verifier(os.path.join(tmp_location, file), size, digest):
			gen_print("%s\tfile '%s' digest failed. requeuing\n" % (cpv, file))
			# previously, we just tried continuing from what we had.
			# doesn't fly.  some mirrors give out shit data, which means the next fetch
			# will do a continuation, and fail digest, thus using up a uri.
			# just wipe the sucker.
			try:	os.unlink(os.path.join(tmp_location, file))
			except OSError:
				gen_print("failed removing %s from tmp dir.  this isn't good!" % file)
			job_queue.publish(entry)
		else:
			if move_required and not debug:
				try:	ret=movefile(os.path.join(tmp_location, file), \
					os.path.join(distdir, file))
				except SystemExit:	raise
				except Exception, e:
					ret = "exception '%s'" % str(e)
					del e
				if not ret or isinstance(ret, str):
					if ret == False:
						ret = "movefile return of False"
					fail_print("%s\tmove failed for file '%s' with %s\n" % (cpv, file, ret))
					continue
				
			try:		new_byte_count += size
			except TypeError: pass
			suc_print("%s\t%s\tadded %i bytes\n" % (cpv, file, size))

	dprint("left job queue looping")

	# to reach here means all queues are empty, and all workers are waiting.
	job_queue.kill_the_workers()

	#bide our time, waiting to absorb all the workers.
	for x in threads:
		dprint("joining %s" % str(x))
		x.join()
		dprint("joined")
	
	#now reclaim the failed logger.
	failed_thread_kill_switch.set()
	failed_thread.join()

	if run_cpidw:
		dprint("collecting cpidw")
		cpidw.join()
		dprint("got it, cpidw")

	# There can be only One...
	# and we're the one, at this point.
	
	end_time = time.time()
	gen_print("\nfinished in %i seconds\n" % int(end_time - start_time))
	gen_print("failed on %i files\n" % failed_thread.failures)
	if cleansing:
		gen_print("cleansed %i files\n" % cleansed_files)
	if death_int:
		gen_print("death watch for %i files\n" % death_watch_count)
	gen_print("new bytes added, %i\n" % new_byte_count)
	sys.exit(exit_code)
